#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
GNSS Module

In Development
"""
from numpy import zeros


def read_sinex_estimate(file):
    """This function reads in the SOLUTION/ESTIMATE block of a SINEX file. It
    returns estimate, a list of tuples:

    estimate = [(code, soln, refEpoch, staX, staY, staZ, staX_sd, staY_sd,
                 staZ_sd[, velX, velY, velZ, velX_sd, velY_sd, velZ_sd])...]

    where:
        * code is the stations's 4-character ID
        * soln is the segment of the stations's time series
        * refEpoch is the epoch of the solution in the form YY:DOY:SSSSS (YY
            is the two digit year, DOY is day of year, and SSSSS is the time of
            day in seconds
        * sta[XYZ] is the station coordinates in the Cartesian reference frame
        * sta[XYZ]_sd is the standard deviation of the station coordinates in
            the Cartesian reference frame
        * vel[XYZ] is the station velocity in the Cartesian reference frame
        * vel[XYZ]_sd is the standard deviation of the station velocity in the
            Cartesian reference frame

    Velocities are not included in all SINEX files and so are only returned if
    present.

    :param file: the input SINEX file
    :return: estimate
    """

    # Create data structures and set variables
    lines = []
    estimate = []
    velocities = False
    go = False
    code = ''
    soln = ''
    epoch = ''
    stax = ''
    stay = ''
    staz = ''
    stax_sd = ''
    stay_sd = ''
    staz_sd = ''
    velx = ''
    vely = ''
    velz = ''
    velx_sd = ''
    vely_sd = ''
    velz_sd = ''

    # Read the SOLUTION/ESTIMATE block into a list and determine if there is
    # any velocity information
    with open(file) as f:
        for line in f:
            if line[:18] == '-SOLUTION/ESTIMATE':
                break
            if go and line[:11] == '*INDEX TYPE':
                pass
            elif go:
                if line[7:10] == 'VEL':
                    velocities = True
                lines.append(line)
            if line[:18] == '+SOLUTION/ESTIMATE':
                go = True

    for line in lines:
        typ = line[7:11]
        if typ == 'STAX':
            code = line[14:18]
            soln = line[23:26].lstrip()
            epoch = line[27:39]
            stax = float(line[47:68])
            stax_sd = float(line[69:80])
        elif typ == 'STAY':
            stay = float(line[47:68])
            stay_sd = float(line[69:80])
        elif typ == 'STAZ':
            staz = float(line[47:68])
            staz_sd = float(line[69:80])
            if not velocities:
                info = (code, soln, epoch, stax, stay, staz, stax_sd, stay_sd,
                        staz_sd)
                estimate.append(info)
        elif typ == 'VELX':
            velx = float(line[47:68])
            velx_sd = float(line[69:80])
        elif typ == 'VELY':
            vely = float(line[47:68])
            vely_sd = float(line[69:80])
        elif typ == 'VELZ':
            velz = float(line[47:68])
            velz_sd = float(line[69:80])
            info = (code, soln, epoch, stax, stay, staz, stax_sd, stay_sd,
                    staz_sd, velx, vely, velz, velx_sd, vely_sd, velz_sd)
            estimate.append(info)

    return estimate


def read_sinex_matrix(file):

    """This function reads in the SOLUTION/MATRIX_ESTIMATE block of a SINEX
    file. It returns matrix, a list of tuples:

    matrix = [(code, soln, var_x, covar_xy, covar_xz, var_y, covar_yz,
            var_z[, var_v_x, covar_v_xy, covar_v_xz, var_v_y, covar_v_yz,
            var_v_z])...]

    where:
        * code is the stations's 4-character ID
        * soln is the segment of the stations's time series
        * var_x is the variance in the X coordinate
        * covar_xy is the covariance between the X and the Y coordinates
        * covar_xz is the covariance between the X and the Z coordinates
        * var_y is the variance in the Y coordinate
        * covar_yz is the covariance between the Y and the Z coordinates
        * var_z is the variance in the Z coordinate
        * var_v_x is the variance in the X velocity
        * covar_v_xy is the covariance between the X and the Y velocities
        * covar_v_xz is the covariance between the X and the Z velocities
        * var_v_y is the variance in the Y velocity
        * covar_v_yz is the covariance between the Y and the Z velocities
        * var_v_z is the variance in the Z velocity

    Velocities are not included in all SINEX files and so their VCV information
    is only returned if they are present.

    :param file: the input SINEX file
    :return: matrix
    """

    # Read in the codes (station names) and solutions, and check for velocities
    data = read_sinex_estimate(file)
    code = []
    soln = []
    velocities = False
    for station in data:
        code.append(station[0])
        soln.append(station[1])
    if len(data[0]) == 15:
        velocities = True

    # Read the SOLUTION/MATRIX_ESTIMATE block into a list and determine if the
    # matrix is upper or lower triangular
    lines = []
    lower_triangular = False
    go = False
    with open(file) as f:
        for line in f:
            if line[:25] == '-SOLUTION/MATRIX_ESTIMATE':
                break
            if go and line[:12] == '*PARA1 PARA2':
                pass
            elif go:
                lines.append(line)
            if line[:25] == '+SOLUTION/MATRIX_ESTIMATE':
                if line[26] == 'L':
                    lower_triangular = True
                go = True

    # Create an array containing the matrix elements
    if velocities:
        n = 6 * int(len(code))
    else:
        n = 3 * int(len(code))
    element = zeros((n, n))
    matrix = []
    for line in lines:
        col = line.rstrip().split()
        for i in range(2, len(col)):
            element[int(col[0]) - 1][int(col[1]) + i - 3] = float(col[i])
    if velocities:
        if lower_triangular:
            for i in range(len(code)):
                info = (code[i], soln[i], element[6 * i][6 * i],
                        element[6 * i + 1][6 * i],
                        element[6 * i + 1][6 * i + 1],
                        element[6 * i + 2][6 * i],
                        element[6 * i + 2][6 * i + 1],
                        element[6 * i + 2][6 * i + 2],
                        element[6 * i + 3][6 * i + 3],
                        element[6 * i + 4][6 * i + 3],
                        element[6 * i + 4][6 * i + 4],
                        element[6 * i + 5][6 * i + 3],
                        element[6 * i + 5][6 * i + 4],
                        element[6 * i + 5][6 * i + 5])
                matrix.append(info)
        else:
            for i in range(len(code)):
                info = (code[i], soln[i], element[6 * i][6 * i],
                        element[6 * i][6 * i + 1], element[6 * i][6 * i + 2],
                        element[6 * i + 1][6 * i + 1],
                        element[6 * i + 1][6 * i + 2],
                        element[6 * i + 2][6 * i + 2],
                        element[6 * i + 3][6 * i + 3],
                        element[6 * i + 3][6 * i + 4],
                        element[6 * i + 3][6 * i + 5],
                        element[6 * i + 4][6 * i + 4],
                        element[6 * i + 4][6 * i + 5],
                        element[6 * i + 5][6 * i + 5])
                matrix.append(info)
    else:
        if lower_triangular:
            for i in range(len(code)):
                info = (code[i], soln[i], element[3 * i][3 * i],
                        element[3 * i + 1][3 * i],
                        element[3 * i + 1][3 * i + 1],
                        element[3 * i + 2][3 * i],
                        element[3 * i + 2][3 * i + 1],
                        element[3 * i + 2][3 * i + 2])
                matrix.append(info)
        else:
            for i in range(len(code)):
                info = (code[i], soln[i], element[3 * i][3 * i],
                        element[3 * i][3 * i + 1], element[3 * i][3 * i + 2],
                        element[3 * i + 1][3 * i + 1],
                        element[3 * i + 1][3 * i + 2],
                        element[3 * i + 2][3 * i + 2])
                matrix.append(info)

    return matrix



#!/usr/bin/env python3

from math import radians, sin, cos, sqrt, atan2, degrees
import numpy as np


def rotation_matrix(lat, lon):
    """Returns the rotation matrix between the local and Cartesian reference
    systems at a given latitude and longitude

    See Section 4.2.3 of the DynAdjust User's Guide v1.0

    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :return: rot_matrix (3x3)
    """
    (rlat, rlon) = (radians(lat), radians(lon))
    rot_matrix = np.array(
        [[-sin(rlon), -sin(rlat) * cos(rlon), cos(rlat) * cos(rlon)],
         [cos(rlon), -sin(rlat) * sin(rlon), cos(rlat) * sin(rlon)],
         [0.0, cos(rlat), sin(rlat)]]
    )
    return rot_matrix


def vcv_cart2local(vcv_cart, lat, lon):
    """Transforms a VCV from the Cartesian to the local reference frame. If
    only a column vector of variances is supplied then they are assumed to be
    uncorrelated and only the column vector of variances is returned.

    See Section 4.4.1 of the DynAdjust User's Guide v1.0

    :param vcv_cart: a VCV in the Cartesian reference frame (3x1 or 3x3)
    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :return: vcv_local (3x1 or 3x3)
    """
    column_vector = False
    if vcv_cart.shape[0] == 3:
        if vcv_cart.shape[1] == 1:
            column_vector = True
            vcv_cart = np.array([[vcv_cart[0, 0], 0.0, 0.0],
                                 [0.0, vcv_cart[1, 0], 0.0],
                                 [0.0, 0.0, vcv_cart[2, 0]]])
        elif vcv_cart.shape[1] == 3:
            pass
        else:
            raise ValueError('Matrix must be either 3x1 or 3x3')
    else:
        raise ValueError('Matrix must be either 3x1 or 3x3')
    rot_matrix = rotation_matrix(lat, lon)
    vcv_local = rot_matrix.transpose() @ vcv_cart @ rot_matrix
    if column_vector:
        vcv_local = np.array([[vcv_local[0, 0]], [vcv_local[1, 1]],
                              [vcv_local[2, 2]]])
    return vcv_local


def vcv_local2cart(vcv_local, lat, lon):
    """Transform a VCV from the local to the Cartesian reference frame. If
    only a column vector of variances is supplied then they are assumed to be
    uncorrelated and only the column vector of variances is returned.

    See Section 4.4.1 of the DynAdjust User's Guide v1.0

    :param vcv_local: a VCV in the local reference frame (3x1 or 3x3)
    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :return: vcv_cart (3x1 or 3x3)
    """
    column_vector = False
    if vcv_local.shape[0] == 3:
        if vcv_local.shape[1] == 1:
            column_vector = True
            vcv_local = np.array([[vcv_local[0, 0], 0.0, 0.0],
                                 [0.0, vcv_local[1, 0], 0.0],
                                 [0.0, 0.0, vcv_local[2, 0]]])
        elif vcv_local.shape[1] == 3:
            pass
        else:
            raise ValueError('Matrix must be either 3x1 or 3x3')
    else:
        raise ValueError('Matrix must be either 3x1 or 3x3')
    rot_matrix = rotation_matrix(lat, lon)
    vcv_cart = rot_matrix @ vcv_local @ rot_matrix.transpose()
    if column_vector:
        vcv_cart = np.array([[vcv_cart[0, 0]], [vcv_cart[1, 1]],
                             [vcv_cart[2, 2]]])
    return vcv_cart


def error_ellipse(vcv):
    """Calculate the semi-major axis, semi-minor axis, and the orientation of
    the error ellipse defined by a VCV

    See Section 7.3.3.1 of the DynaNet User's Guide v3.3

    :param vcv: a VCV (3x3)
    :return: a, semi-major axis
    :return: b, semi-minor axis
    :return: orientation, the orientation of the error ellipse
    """
    z = sqrt((vcv[0, 0] - vcv[1, 1])**2 + 4 * vcv[0, 1]**2)
    a = sqrt(0.5 * (vcv[0, 0] + vcv[1, 1] + z))
    b = sqrt(0.5 * (vcv[0, 0] + vcv[1, 1] - z))
    orientation = 90 - degrees(0.5 * atan2((2 * vcv[0, 1]),
                                           (vcv[0, 0] - vcv[1, 1])))

    return a, b, orientation


def relative_error(lat, lon, var1, var2, cov12):
    """
    Function to compute relative error between two 3D stations:
        - 2D relative error ellipse [semi-major axis, semi-minor axis, bearing]
        - 1D relative 'up' error
        
    Adapted from Harvey B.R. (1998) Practical least squares and statistics for surveyors,
    Monograph 13, Section 4.4, p.135

    :param lat: latitude at Stn1 in decimal degrees
    :param lon: longitude at Stn1 in decimal degrees
    :param var1: 3x3 Cartesian XYZ variance matrix of Stn1
    :param var2: 3x3 Cartesian XYZ variance matrix of Stn2
    :param cov12: 3x3 Cartesian XYZ covariance block between Stn1 and Stn2
    :return: Relative error ellipse components [smaj, smin, brg] and relative 'up' error
    """

    # rotate matrices from cartesian to local
    var1_enu = vcv_cart2local(var1, lat, lon)
    var2_enu = vcv_cart2local(var2, lat, lon)
    cov12_enu = vcv_cart2local(cov12, lat, lon)

    # form relative variance matrix
    rel_var = np.zeros((3, 3))

    # diagonal terms
    rel_var[0, 0] = var1_enu[0, 0] + var2_enu[0, 0] - 2 * cov12_enu[0, 0]
    rel_var[1, 1] = var1_enu[1, 1] + var2_enu[1, 1] - 2 * cov12_enu[1, 1] 
    rel_var[2, 2] = var1_enu[2, 2] + var2_enu[2, 2] - 2 * cov12_enu[2, 2]

    # east-north covariance
    rel_var[0, 1] = var1_enu[0, 1] + var2_enu[0, 1] - cov12_enu[0, 1] - cov12_enu[1, 0]
    rel_var[1, 0] = rel_var[0, 1]

    # east-up covariance
    rel_var[0, 2] = var1_enu[0, 2] + var2_enu[0, 2] - cov12_enu[0, 2] - cov12_enu[2, 0]
    rel_var[2, 0] = rel_var[0, 2]

    # north-up covariance
    rel_var[1, 2] = var1_enu[1, 2] + var2_enu[1, 2] - cov12_enu[1, 2] - cov12_enu[2, 1]
    rel_var[2, 1] = rel_var[1, 2]

    # relative error ellipse [smaj, smin, brg]
    ree = error_ellipse(rel_var)

    # relative up error
    rue = rel_var[2, 2] ** 0.5

    return ree[0], ree[1], ree[2], rue


def circ_hz_pu(a, b):
    """Calculate the circularised horizontal PU from the semi-major and
    semi-minor axes of an error ellipse

    :param a: semi-major axis
    :param b: semi-minor axis
    :return: r the radius of the circularised error
    """
    q0 = 1.960790
    q1 = 0.004071
    q2 = 0.114276
    q3 = 0.371625
    c = b / a
    k = q0 + q1 * c + q2 * c**2 + q3 * c**3
    r = a * k

    return r


def k_val95(dof):
    """
    Returns the Coverage Factor k for a given 1 sigma (68.27%) Standard
    Deviation to allow conversion to a 95% Standard Deviation. This
    uses a simplified table of k values rounded to 5 decimal places
    for Degrees of Freedom (DOF) in the range 1 to 120. For DOF above
    120, returns k value of 1.96 and for DOF below 1, returns k value
    for DOF = 1. Coverage Factor produced using following scipy stats
    code: stats.t.ppf(1-0.025,dof)

    :param dof: Degrees of Freedom (Number of Measurements Minus 1)
    :return: Coverage Factor k
    """
    if not isinstance(dof, int):
        raise TypeError('Degrees of Freedom must be Int')
    if dof < 1:
        return ttable_p95[0]
    elif dof > 120:
        return 1.96
    else:
        return ttable_p95[dof - 1]


ttable_p95 = ([12.7062, 4.30265, 3.18245, 2.77645, 2.57058, 2.44691,
              2.36462, 2.30600, 2.26216, 2.22814, 2.20099, 2.17881,
              2.16037, 2.14479, 2.13145, 2.11991, 2.10982, 2.10092,
              2.09302, 2.08596, 2.07961, 2.07387, 2.06866, 2.06390,
              2.05954, 2.05553, 2.05183, 2.04841, 2.04523, 2.04227,
              2.03951, 2.03693, 2.03452, 2.03224, 2.03011, 2.02809,
              2.02619, 2.02439, 2.02269, 2.02108, 2.01954, 2.01808,
              2.01669, 2.01537, 2.01410, 2.01290, 2.01174, 2.01063,
              2.00958, 2.00856, 2.00758, 2.00665, 2.00575, 2.00488,
              2.00404, 2.00324, 2.00247, 2.00172, 2.00100, 2.00030,
              1.99962, 1.99897, 1.99834, 1.99773, 1.99714, 1.99656,
              1.99601, 1.99547, 1.99495, 1.99444, 1.99394, 1.99346,
              1.99300, 1.99254, 1.99210, 1.99167, 1.99125, 1.99085,
              1.99045, 1.99006, 1.98969, 1.98932, 1.98896, 1.98861,
              1.98827, 1.98793, 1.98761, 1.98729, 1.98698, 1.98667,
              1.98638, 1.98609, 1.98580, 1.98552, 1.98525, 1.98498,
              1.98472, 1.98447, 1.98422, 1.98397, 1.98373, 1.98350,
              1.98326, 1.98304, 1.98282, 1.98260, 1.98238, 1.98217,
              1.98197, 1.98177, 1.98157, 1.98137, 1.98118, 1.98099,
              1.98081, 1.98063, 1.98045, 1.98027, 1.98010, 1.97993])



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Geodesy Module
"""

from math import degrees, radians, sqrt, sin, cos, tan, asin, acos, atan, atan2
import numpy as np
from geodepy.constants import grs80, utm
from geodepy.convert import geo2grid, grid2geo, angular_typecheck
from geodepy.statistics import rotation_matrix
from geodepy.survey import radiations


def enu2xyz(lat, lon, east, north, up):
    """Convert a column vector in the local reference frame to a column vector
    in the Cartesian reference frame.
    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :param east: in metres
    :param north: in metres
    :param up: in metres
    :return: x, y, z in metres
    """
    rot_matrix = rotation_matrix(angular_typecheck(lat), angular_typecheck(lon))
    enu = np.array([[east], [north], [up]])
    xyz = rot_matrix @ enu
    x = xyz[0, 0]
    y = xyz[1, 0]
    z = xyz[2, 0]
    return x, y, z


def xyz2enu(lat, lon, x, y, z):
    """Convert a column vector in the Cartesian reference frame to a column
    vector in the local reference frame.
    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :param x: in metres
    :param y: in metres
    :param z: in metres
    :return: east, north, up in metres
    """
    rot_matrix = rotation_matrix(angular_typecheck(lat), angular_typecheck(lon))
    xyz = np.array([[x], [y], [z]])
    enu = rot_matrix.transpose() @ xyz
    east = enu[0, 0]
    north = enu[1, 0]
    up = enu[2, 0]
    return east, north, up


def vincdir(lat1, lon1, azimuth1to2, ell_dist, ellipsoid=grs80):
    """
    Vincenty's Direct Formula
    :param lat1: Latitude of Point 1 (decimal degrees)
    :type lat1: float (decimal degrees), DMSAngle or DDMAngle
    :param lon1: Longitude of Point 1 (decimal degrees)
    :type lon1: float (decimal degrees), DMSAngle or DDMAngle
    :param azimuth1to2: Azimuth from Point 1 to 2 (decimal degrees)
    :type azimuth1to2: float (decimal degrees), DMSAngle or DDMAngle
    :param ell_dist: Ellipsoidal Distance between Points 1 and 2 (metres)
    :param ellipsoid: Ellipsoid Object
    :return: lat2: Latitude of Point 2 (Decimal Degrees),
             lon2: Longitude of Point 2 (Decimal Degrees),
             azimuth2to1: Azimuth from Point 2 to 1 (Decimal Degrees)

    Code review: 14-08-2018 Craig Harrison
    """

    # Convert Angles to Decimal Degrees (if required)
    lat1 = angular_typecheck(lat1)
    lon1 = angular_typecheck(lon1)
    azimuth1to2 = radians(angular_typecheck(azimuth1to2))

    # Equation numbering is from the GDA2020 Tech Manual v1.0

    # Eq. 88
    u1 = atan((1 - ellipsoid.f) * tan(radians(lat1)))

    # Eq. 89
    sigma1 = atan2(tan(u1), cos(azimuth1to2))

    # Eq. 90
    alpha = asin(cos(u1) * sin(azimuth1to2))

    # Eq. 91
    u_squared = cos(alpha)**2 \
        * (ellipsoid.semimaj**2 - ellipsoid.semimin**2) \
        / ellipsoid.semimin**2

    # Eq. 92
    a = 1 + (u_squared / 16384) \
        * (4096 + u_squared * (-768 + u_squared * (320 - 175 * u_squared)))

    # Eq. 93
    b = (u_squared / 1024) \
        * (256 + u_squared * (-128 + u_squared * (74 - 47 * u_squared)))

    # Eq. 94
    sigma = ell_dist / (ellipsoid.semimin * a)

    # Iterate until the change in sigma, delta_sigma, is insignificant (< 1e-9)
    # or after 1000 iterations have been completed
    two_sigma_m = 0
    for i in range(1000):

        # Eq. 95
        two_sigma_m = 2*sigma1 + sigma

        # Eq. 96
        delta_sigma = b * sin(sigma) * (cos(two_sigma_m) + (b/4)
                                        * (cos(sigma)
                                           * (-1 + 2 * cos(two_sigma_m)**2)
                                           - (b/6) * cos(two_sigma_m)
                                           * (-3 + 4 * sin(sigma)**2)
                                           * (-3 + 4 * cos(two_sigma_m)**2)))
        new_sigma = (ell_dist / (ellipsoid.semimin * a)) + delta_sigma
        sigma_change = new_sigma - sigma
        sigma = new_sigma

        if abs(sigma_change) < 1e-12:
            break

    # Calculate the Latitude of Point 2
    # Eq. 98
    lat2 = atan2(sin(u1)*cos(sigma) + cos(u1)*sin(sigma)*cos(azimuth1to2),
                 (1 - ellipsoid.f)
                 * sqrt(sin(alpha)**2 + (sin(u1)*sin(sigma)
                        - cos(u1)*cos(sigma)*cos(azimuth1to2))**2))
    lat2 = degrees(lat2)

    # Calculate the Longitude of Point 2
    # Eq. 99
    lon = atan2(sin(sigma)*sin(azimuth1to2),
                cos(u1)*cos(sigma) - sin(u1)*sin(sigma)*cos(azimuth1to2))

    # Eq. 100
    c = (ellipsoid.f/16)*cos(alpha)**2 \
        * (4 + ellipsoid.f*(4 - 3*cos(alpha)**2))

    # Eq. 101
    omega = lon - (1-c)*ellipsoid.f*sin(alpha) \
        * (sigma + c*sin(sigma)*(cos(two_sigma_m) + c*cos(sigma)
                                 * (-1 + 2*cos(two_sigma_m)**2)))

    # Eq. 102
    lon2 = float(lon1) + degrees(omega)

    # Calculate the Reverse Azimuth
    azimuth2to1 = degrees(atan2(sin(alpha), -sin(u1)*sin(sigma)
                          + cos(u1)*cos(sigma)*cos(azimuth1to2))) + 180

    return round(lat2, 11), round(lon2, 11), round(azimuth2to1, 9)


def vincinv(lat1, lon1, lat2, lon2, ellipsoid=grs80):
    """
    Vincenty's Inverse Formula
    :param lat1: Latitude of Point 1 (decimal degrees)
    :type lat1: float (decimal degrees), DMSAngle or DDMAngle
    :param lon1: Longitude of Point 1 (decimal degrees)
    :type lon1: float (decimal degrees), DMSAngle or DDMAngle
    :param lat2: Latitude of Point 2 (decimal degrees)
    :type lat2: float (decimal degrees), DMSAngle or DDMAngle
    :param lon2: Longitude of Point 2 (decimal degrees)
    :type lon2: float (decimal degrees), DMSAngle or DDMAngle
    :param ellipsoid: Ellipsoid Object
    :return: ell_dist: Ellipsoidal Distance between Points 1 and 2 (m),
             azimuth1to2: Azimuth from Point 1 to 2 (Decimal Degrees),
             azimuth2to1: Azimuth from Point 2 to 1 (Decimal Degrees)
    Code review: 14-08-2018 Craig Harrison
    """

    # Convert Angles to Decimal Degrees (if required)
    lat1 = angular_typecheck(lat1)
    lon1 = angular_typecheck(lon1)
    lat2 = angular_typecheck(lat2)
    lon2 = angular_typecheck(lon2)

    # Exit if the two input points are the same
    if lat1 == lat2 and lon1 == lon2:
        return 0, 0, 0

    # Equation numbering is from the GDA2020 Tech Manual v1.0

    # Eq. 71
    u1 = atan((1 - ellipsoid.f) * tan(radians(lat1)))

    # Eq. 72
    u2 = atan((1 - ellipsoid.f) * tan(radians(lat2)))

    # Eq. 73; initial approximation
    lon = radians(lon2 - lon1)
    omega = lon

    # Iterate until the change in lambda, lambda_sigma, is insignificant
    # (< 1e-12) or after 1000 iterations have been completed
    alpha = 0
    sigma = 0
    cos_two_sigma_m = 0
    for i in range(1000):

        # Eq. 74
        sin_sigma = sqrt((cos(u2)*sin(lon))**2
                         + (cos(u1)*sin(u2) - sin(u1)*cos(u2)*cos(lon))**2)

        # Eq. 75
        cos_sigma = sin(u1)*sin(u2) + cos(u1)*cos(u2)*cos(lon)

        # Eq. 76
        sigma = atan2(sin_sigma, cos_sigma)

        # Eq. 77
        alpha = asin((cos(u1)*cos(u2)*sin(lon)) / sin_sigma)

        # Eq. 78
        cos_two_sigma_m = cos(sigma) - (2*sin(u1)*sin(u2) / cos(alpha)**2)

        # Eq. 79
        c = (ellipsoid.f / 16) * cos(alpha)**2 * (4 + ellipsoid.f
                                                  * (4 - 3*cos(alpha)**2))

        # Eq. 80
        new_lon = omega + (1 - c) * ellipsoid.f * sin(alpha) * (sigma + c*sin(sigma)
                                                                * (cos_two_sigma_m + c * cos(sigma)
                                                                   * (-1 + 2*(cos_two_sigma_m**2))))

        delta_lon = new_lon - lon
        lon = new_lon

        if abs(delta_lon) < 1e-12:
            break

    # Eq. 81
    u_squared = cos(alpha)**2 \
        * (ellipsoid.semimaj**2 - ellipsoid.semimin**2) \
        / ellipsoid.semimin**2

    # Eq. 82
    a = 1 + (u_squared / 16384) \
        * (4096 + u_squared * (-768 + u_squared * (320 - 175 * u_squared)))

    # Eq. 83
    b = (u_squared / 1024) \
        * (256 + u_squared * (-128 + u_squared * (74 - 47 * u_squared)))

    # Eq. 84
    delta_sigma = b*sin(sigma) * (cos_two_sigma_m + (b / 4)
                                  * (cos(sigma) * (-1 + 2*cos_two_sigma_m**2)
                                  - (b / 6)*cos_two_sigma_m
                                  * (-3 + 4*sin(sigma)**2)
                                  * (-3 + 4*cos_two_sigma_m**2)))

    # Calculate the ellipsoidal distance
    # Eq. 85
    ell_dist = ellipsoid.semimin*a * (sigma - delta_sigma)

    # Calculate the azimuth from point 1 to point 2
    azimuth1to2 = degrees(atan2((cos(u2)*sin(lon)),
                                (cos(u1)*sin(u2)
                                 - sin(u1)*cos(u2)*cos(lon))))

    if azimuth1to2 < 0:
        azimuth1to2 = azimuth1to2 + 360

    # Calculate the azimuth from point 2 to point 1
    azimuth2to1 = degrees(atan2(cos(u1)*sin(lon),
                                (-sin(u1)*cos(u2)
                                 + cos(u1)*sin(u2)*cos(lon)))) + 180

    # Meridian Critical Case Tests
    #if lon1 == lon2 and lat1 > lat2:
    #    azimuth1to2 = 180
    #    azimuth2to1 = 0
    #elif lon1 == lon2 and lat1 < lat2:
    #    azimuth1to2 = 0
    #    azimuth2to1 = 180

    return round(ell_dist, 3), round(azimuth1to2, 9), round(azimuth2to1, 9)


def vincdir_utm(zone1, east1, north1, grid1to2, grid_dist,
                hemisphere='south', ellipsoid=grs80):
    """
    Perform Vincenty's Direct Computation using UTM Grid Coordinates, a
    grid bearing and grid distance.
    Note: Point 2 UTM Coordinates use the Zone specified for Point 1, even if
    Point 2 would typically be computed in a different zone. This keeps the grid
    bearings and line scale factor all relative to the same UTM Zone.
    :param zone1: Point 1 Zone Number - 1 to 60
    :param east1: Point 1 Easting (m, within 3330km of Central Meridian)
    :param north1: Point 1 Northing (m, 0 to 10,000,000m)
    :param grid1to2: Grid Bearing from Point 1 to 2 (decimal degrees),
    :param grid_dist: UTM Grid Distance between Points 1 and 2 (m)
    :param hemisphere: String - 'North' or 'South'(default)
    :param ellipsoid: Ellipsoid Object (default: GRS80)
    :return: zone2: Point 2 Zone Number - 1 to 60
             east2: Point 2 Easting (m, within 3330km of Central Meridian)
             north2: Point 2 Northing (m, 0 to 10,000,000m)
             grid2to1: Grid Bearing from Point 2 to 1 (decimal degrees)
             lsf: Line Scale Factor (for Point 1 Zone)
    """
    # Convert angle to decimal degrees (if required)
    grid1to2 = angular_typecheck(grid1to2)

    # Convert UTM Coords to Geographic
    lat1, lon1, psf1, gridconv1 = grid2geo(zone1, east1, north1,
                                           hemisphere, ellipsoid)

    # Convert Grid Bearing to Geodetic Azimuth
    az1to2 = grid1to2 - gridconv1

    # Estimate Line Scale Factor (LSF)
    zone2, east2, north2 = (zone1, *radiations(east1, north1,
                                               grid1to2, grid_dist))
    lsf = line_sf(zone1, east1, north1, zone2, east2, north2)

    # Iteratively estimate Pt 2 Coordinates, refining LSF each time
    lsf_diff = 1
    max_iter = 10
    while lsf_diff > 1e-9:
        lsf_previous = lsf
        lat2, lon2, az2to1 = vincdir(lat1, lon1, az1to2,
                                     grid_dist / lsf, ellipsoid)
        (hemisphere2, zone2, east2,
         north2, psf2, gridconv2) = geo2grid(lat2, lon2,
                                             zone1, ellipsoid)
        lsf = line_sf(zone1, east1, north1,
                      zone2, east2, north2,
                      hemisphere, ellipsoid)
        lsf_diff = abs(lsf_previous - lsf)

    # Compute Grid Bearing for Station 2
    lat2, lon2, psf2, gridconv2 = grid2geo(zone2, east2, north2,
                                           hemisphere, ellipsoid)
    grid2to1 = az2to1 + gridconv2

    return zone2, east2, north2, grid2to1, lsf


def vincinv_utm(zone1, east1, north1, zone2, east2, north2,
                hemisphere='south', ellipsoid=grs80):
    """
    Perform Vincentys Inverse Computation using UTM Grid Coordinates
    Note: Where coordinates from different zones are used, UTM Grid Distance
    is relative to Point 1's Zone. Grid Bearings of Points 1 and 2 are
    relative to each of their respective Zones.
    :param zone1: Point 1 Zone Number - 1 to 60
    :param east1: Point 1 Easting (m, within 3330km of Central Meridian)
    :param north1: Point 1 Northing (m, 0 to 10,000,000m)
    :param zone2: Point 2 Zone Number - 1 to 60
    :param east2: Point 2 Easting (m, within 3330km of Central Meridian)
    :param north2: Point 2 Northing (m, 0 to 10,000,000m)
    :param hemisphere: String - 'North' or 'South'(default)
    :param ellipsoid: Ellipsoid Object (default: GRS80)
    :return: grid_dist: UTM Grid Distance between Points 1 and 2 (m),
             grid1to2: Grid Bearing from Point 1 to 2 (decimal degrees),
             grid2to1: Grid Bearing from Point 2 to 1 (decimal degrees)
             lsf: Line Scale Factor (for Point 1 Zone)
    """
    # Convert utm to geographic
    pt1 = grid2geo(zone1, east1, north1, hemisphere, ellipsoid)
    pt2 = grid2geo(zone2, east2, north2, hemisphere, ellipsoid)
    ell_dist, az1to2, az2to1 = vincinv(pt1[0], pt1[1],
                                       pt2[0], pt2[1], ellipsoid)

    # Compute Grid Distance using Line Scale Factor
    lsf = line_sf(zone1, east1, north1,
                  zone2, east2, north2, hemisphere, ellipsoid)
    grid_dist = ell_dist * lsf

    # Compute Grid Bearings using Grid Convergence
    grid1to2 = az1to2 + pt1[3]
    grid2to1 = az2to1 + pt2[3]

    return grid_dist, grid1to2, grid2to1, lsf


def line_sf(zone1, east1, north1, zone2, east2, north2,
            hemisphere='south', ellipsoid=grs80, projection=utm):
    """
    Computes Line Scale Factor for a pair of Transverse Mercator Coordinates
    Ref: Deakin 2010, Traverse Computations on the Ellipsoid and on the
    Universal Transverse Mercator Projection, pp 35
    http://www.mygeodesy.id.au/documents/Trav_Comp_V2.1.pdf
    :param zone1: Station 1 Zone Number - 1 to 60
    :param east1: Station 1 Easting (m, within 3330km of Central Meridian)
    :param north1: Station 1 Northing (m, 0 to 10,000,000m)
    :param zone2: Station 2 Zone Number - 1 to 60
    :param east2: Station 2 Easting (m, within 3330km of Central Meridian)
    :param north2: Station 2 Northing (m, 0 to 10,000,000m)
    :param hemisphere: String - 'North' or 'South'(default)
    :param ellipsoid: Ellipsoid Object (default GRS80)
    :param projection: Projection Object (default Universal Transverse Mercator)
    :return: Line Scale Factor (relative to Zone 1 if different zones
    are entered)
    """
    # Re-project cross-zone coordinate to same UTM zone
    if zone1 != zone2:
        # Re-project Station 2 Coordinates to same zone as Station 1
        stn2_geo = grid2geo(zone2, east2, north2, hemisphere, ellipsoid)
        stn2_zone1 = geo2grid(stn2_geo[0], stn2_geo[1], zone1, ellipsoid)
        zone2 = stn2_zone1[1]
        east2 = stn2_zone1[2]
        north2 = stn2_zone1[3]

    # Comute easting distances from Central Meridian
    eastofcm1 = east1 - projection.falseeast
    eastofcm2 = east2 - projection.falseeast

    # Compute Mean Latitude
    lat1 = grid2geo(zone1, east1, north1, hemisphere, ellipsoid)[0]
    lat2 = grid2geo(zone2, east2, north2, hemisphere, ellipsoid)[0]
    lat_mean = (lat1 + lat2) / 2

    # Compute Line Scale Factor (Deakin 2010 Eq. 13)
    r_sq_m = (rho(lat_mean, ellipsoid) *
              nu(lat_mean, ellipsoid) *
              projection.cmscale ** 2)
    k1 = ((eastofcm1 ** 2 + eastofcm1 * eastofcm2 + eastofcm2 ** 2) /
          (6 * r_sq_m))
    k2 = ((eastofcm1 ** 2 + eastofcm1 * eastofcm2 + eastofcm2 ** 2) /
          (36 * r_sq_m))
    return projection.cmscale * (1 + k1 * (1 + k2))


def rho(lat, ellipsoid=grs80):
    """
    Return the radius of curvature of the ellipsoid in the meridian plane
    (rho) at a given latitude
    :param lat: latitude in decimal degrees
    :param ellipsoid: Ellipsoid Object
    :return: rho at specified latitude
    """
    return ((ellipsoid.semimaj * (1 - ellipsoid.ecc1sq)) /
            (1 - ellipsoid.ecc1sq * (sin(radians(lat)) ** 2)) ** 1.5)


def nu(lat, ellipsoid=grs80):
    """
    Return the radius of curvature of the ellipsoid in the prime vertical plane
    (nu) at a given latitude
    :param lat: latitude in decimal degrees
    :param ellipsoid: Ellipsoid Object
    :return: nu at specified latitude
    """
    return (ellipsoid.semimaj /
            sqrt(1 - ellipsoid.ecc1sq * (sin(radians(lat)) ** 2)))



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
FileIO Module
"""

from geodepy.convert import hp2dec


class DNACoord(object):
    def __init__(self, pointid, const, easting, northing, zone, lat,
                 long, ortho_ht, ell_ht, x, y, z, x_sd, y_sd, z_sd, desc):
        self.pointid = pointid
        self.const = const
        self.easting = easting
        self.northing = northing
        self.zone = zone
        self.lat = lat
        self.long = long
        self.ortho_ht = ortho_ht
        self.ell_ht = ell_ht
        self.x = x
        self.y = y
        self.z = z
        self.x_sd = x_sd
        self.y_sd = y_sd
        self.z_sd = z_sd
        self.desc = desc

    def converthptodd(self):
        self.lat = hp2dec(self.lat)
        self.long = hp2dec(self.long)


def read_dnacoord(fn):
    coord_list = []
    with open(fn, 'r') as file:
        dnadata = file.readlines()
        for line in dnadata:
            pointid = line[0:20]
            const = line[21:25]
            easting = float(line[28:40])
            northing = float(line[41:58])
            zone = int(line[60:63])
            lat = float(line[63:78])
            long = float(line[78:92])
            ortho_ht = float(line[93:103])
            ell_ht = float(line[103:114])
            x = float(line[115:129])
            y = float(line[130:144])
            z = float(line[145:159])
            x_sd = float(line[160:171])
            y_sd = float(line[172:181])
            z_sd = float(line[182:191])
            desc = line[192:-1]
            record = DNACoord(pointid.strip(), const.strip(), easting,
                              northing, zone, lat, long, ortho_ht, ell_ht, x,
                              y, z, x_sd, y_sd, z_sd, desc.strip())
            coord_list.append(record)
    return coord_list




__all__ = ["grid2geoio", "geo2gridio", "gdatrans7"]

import geodepy.convert
import pandas as pd
import tkinter as ttk
import geodepy.transform as tf
import geodepy.convert as cv
import geodepy.constants as cs

'''
The inputoutput module acts as the backend for the GUI and manages the calls
to the GeodePy functions. The functions are rewritten  from the transform
module but use pandas to improve efficiency. Functions placed here to not
overcrowd the transform module.
'''


def grid2geoio(fn, fn_out, easting, northing, utmzone, geotypeout):
    """
        Input:
         The CSV data must have headers.

         Inputs passed the the function:
         :param fn: input file path
         :param fn_out: file output path
         :param easting: is the column name in the csv where the eastings are stored
         :param northing: is the column name in the csv where the northings are stored
         :param utmzone: is the column name in the csv where the UTM Zone is stored
         :param geotypeout: format of latitude and longitude output e.g. DD or DMS

        Output:
        Data in the file is output to the fn_out location as a CSV.
        """

    # Check whether geotypeout value is from GUI or if function called within other code, then check that a valid
    # geotypeout value was provided
    if isinstance(geotypeout, ttk.StringVar):
        geotypeout = geotypeout.get()
    geotypeout_options = ["DD", "DMS"]
    if geotypeout not in geotypeout_options:
        raise ValueError("Invalid geotypeout. Expected one of: %s" % geotypeout_options)

    # Opens the CSV as a pandas dataframe
    csvdf = pd.read_csv(fn, low_memory=False)
    # Sends data to tf.grid2geo
    returned = csvdf.apply(lambda x: geodepy.convert.grid2geo(x[utmzone], x[easting], x[northing]), axis=1)
    # Convert tuple returned to DafaFrame
    resultdf = pd.DataFrame(list(returned), columns=['Latitude', 'Longitude', 'Point Scale Factor', 'Grid Convergence'])
    if geotypeout == 'DMS':
        resultdf['Latitude'] = resultdf['Latitude'].apply(cv.dec2hp)
        resultdf['Longitude'] = resultdf['Longitude'].apply(cv.dec2hp)
    # Adds the results to the original dataframe from the csv
    csvdf = pd.concat([csvdf, resultdf], axis=1)
    # Writes the dataframe to a csv
    csvdf.to_csv(fn_out, index=False)


def geo2gridio(fn, fn_out, latitude, longitude, geotypein):
    """
        Input:
        The CSV data must have headers.

         Inputs passed the the function:
         :param fn: input file path
         :param fn_out: file output path
         :param latitude: is the column name in the csv where the latitudes are stored
         :param longitude: is the column name in the csv where the longitudes are stored
         :param geotypein: format of latitude and longitude e.g. DD or DMS

        Output:
        Data in the file is output to the fn_out location as a CSV.
        """

    # Check whether geotypein value is from GUI or if function called within other code, then check that a valid
    # geotypein value was provided
    if isinstance(geotypein, ttk.StringVar):
        geotypein = geotypein.get()
    geotypein_options = ["DD", "DMS"]
    if geotypein not in geotypein_options:
        raise ValueError("Invalid geotypein. Expected one of: %s" % geotypein_options)

    # Opens the CSV as a pandas dataframe
    csvdf = pd.read_csv(fn, low_memory=False)

    # Converts DMS lat and long to Decimal Degrees if required
    if geotypein == 'DMS':
        csvdf['LatitudeDD'] = csvdf[latitude].apply(cv.hp2dec)
        csvdf['LongitudeDD'] = csvdf[longitude].apply(cv.hp2dec)
        latitude = 'LatitudeDD'
        longitude = 'LongitudeDD'

    # Sends data to tf.geo2grid
    returned = csvdf.apply(lambda x: geodepy.convert.geo2grid(x[latitude], x[longitude]), axis=1)
    # Convert tuple returned from tf.gepo2grid into DafaFrame
    resultdf = pd.DataFrame(list(returned), columns=['Hemisphere', 'UTMZone', 'Easting', 'Northing',
                                                     'Point Scale Factor', 'Grid Convergence'])
    # Adds the results to the original dataframe from the csv
    csvdf = pd.concat([csvdf, resultdf], axis=1)
    # Writes the dataframe to a csv
    csvdf.to_csv(fn_out, index=False)


def gdatrans7(fn, fn_out, latitude, longitude, ellht, gdageotypein, direction):

    """
    Input:
     The CSV data must have headers.

     Inputs passed the the function:
     :param fn: input file path
     :param fn_out: file output path
     :param latitude: is the column name in the csv where the latitudes are stored
     :param longitude: is the column name in the csv where the longitudes are stored
     :param ellht: is the column name in the csv where the ellipsoidal heights are stored
     :param gdageotypein: format of latitude and longitude e.g. DD or DMS
     :param direction: either "94to2020" or "2020to94". Specifies the datum to transform from and to.

     trans: Transformation parameter from the geodepy.constants module. The trans parameters are hard coded below.
     Parameters were sourced from GDA2020 Technical Manual. Need to work on a way of entering and passing different
     transformation parameters to the function.

    Output:
    Data in the file is output to the fn_out location as a CSV.
    """

    # Check whether direction value is from GUI or if function called within other code, then check that a valid
    # direction value was provided
    if isinstance(direction, ttk.StringVar):
        direction = direction.get()
    direction_options = ["94to2020", "2020to94"]
    if direction not in direction_options:
        raise ValueError("Invalid direction. Expected one of: %s" % direction_options)

    # Check whether gdageotypein value is from GUI or if function called within other code, then check that a valid
    # gdageotypein value was provided
    if isinstance(gdageotypein, ttk.StringVar):
        gdageotypein = gdageotypein.get()
    gdageotypein_options = ["DD", "DMS"]
    if gdageotypein not in gdageotypein_options:
        raise ValueError("Invalid gdageotypein. Expected one of: %s" % gdageotypein_options)

    # Sets up the direction of the transformation and sets up the output column naming
    if direction == "94to2020":
        trans = cs.Transformation('MGA94', 'MGA2020', '1994', 0.06155, -0.01087, -0.04019, -0.009994, -0.0394924,
                                  -0.0327221, -0.0328979)
        translat = "GDA2020Latitude"
        translong = "GDA2020Longitude"
        transellht = "GDA2020EllHt"
    else:
        trans = cs.Transformation('MGA94', 'MGA2020', '1994', 0.06155, -0.01087, -0.04019, -0.009994, -0.0394924,
                                  -0.0327221, -0.0328979).__neg__()
        translat = "GDA1994Latitude"
        translong = "GDA1994Longitude"
        transellht = "GDA1994EllHt"

    # Opens the CSV as a pandas dataframe
    csvdf = pd.read_csv(fn, low_memory=False)

    # Converts DMS lat and long to Decimal Degrees if required
    if gdageotypein == 'DMS':
        csvdf['LatitudeDMS'] = csvdf[latitude]
        csvdf['LongitudeDMS'] = csvdf[longitude]
        csvdf['Latitude'] = csvdf[latitude].apply(cv.hp2dec)
        csvdf['Longitude'] = csvdf[longitude].apply(cv.hp2dec)

    # Converts Lat, Long & ElipHt XYZ
    returned = csvdf.apply(lambda x: geodepy.convert.llh2xyz(x[latitude], x[longitude], x[ellht]), axis=1)
    # Converts the results from llh2xyz into a dataframe
    xyzresultdf = pd.DataFrame(list(returned), columns=['CartesianX', 'CartesianY', 'CartesianZ'])

    # Combines the results with the original dataframe
    csvdf = pd.concat([csvdf, xyzresultdf], axis=1)

    # Transforms Lat, Long & ElipHt XYZ using a 7 parameter transformation
    transformed = csvdf.apply(lambda x: tf.conform7(x.CartesianX, x.CartesianY, x.CartesianZ, trans), axis=1)
    # Converts the results from conform7 into a dataframe
    transresultdf = pd.DataFrame(list(transformed), columns=['TransCartesianX', 'TransCartesianY', 'TransCartesianZ'])

    # Combines the results with the original dataframe
    csvdf = pd.concat([csvdf, transresultdf], axis=1)

    # Converts the the transformed XYZ coordinates back to Lats, Longs and Ellipsoidal Heights
    returned = csvdf.apply(lambda x: geodepy.convert.xyz2llh(x.TransCartesianX, x.TransCartesianY, x.TransCartesianZ), axis=1)
    # Converts the results from llh2xyz into a dataframe
    llhresultdf = pd.DataFrame(list(returned), columns=[translat, translong, transellht])

    # Combines the results with the original dataframe
    csvdf = pd.concat([csvdf, llhresultdf], axis=1)
    csvdf = csvdf.drop(['CartesianX', 'CartesianY', 'CartesianZ', 'TransCartesianX', 'TransCartesianY',
                        'TransCartesianZ'], axis=1)

    # Writes the csv dataframe to a csv file
    csvdf.to_csv(fn_out, index=False)



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Transform Module

Ref1
http://www.icsm.gov.au/sites/default/files/GDA2020TechnicalManualV1.1.1.pdf

Ref2
http://www.mygeodesy.id.au/documents/Karney-Krueger%20equations.pdf
"""

import datetime
from math import radians
import numpy as np
from geodepy.constants import Transformation, atrf_gda2020,\
    gda94_to_gda2020
from geodepy.convert import hp2dec, geo2grid, \
    grid2geo, xyz2llh, llh2xyz


def conform7(x, y, z, trans):
    """
    Performs a Helmert 7 Parameter Conformal Transformation using Cartesian point co-ordinates
    and a predefined transformation object.
    :param x: Cartesian X (m)
    :param y: Cartesian Y (m)
    :param z: Cartesian Z (m)
    :param trans: Transformation Object (note: this function ignores all time-dependent variables)
    :return: Transformed X, Y, Z Cartesian Co-ordinates
    """
    if type(trans) != Transformation:
        raise ValueError('trans must be a Transformation Object')
    # Create XYZ Vector
    xyz_before = np.array([[x],
                           [y],
                           [z]])
    # Convert Units for Transformation Parameters
    scale = trans.sc / 1000000
    rx = radians(hp2dec(trans.rx / 10000))
    ry = radians(hp2dec(trans.ry / 10000))
    rz = radians(hp2dec(trans.rz / 10000))
    # Create Translation Vector
    translation = np.array([[trans.tx],
                            [trans.ty],
                            [trans.tz]])
    # Create Rotation Matrix
    rotation = np.array([[1., rz, -ry],
                         [-rz, 1., rx],
                         [ry, -rx, 1.]])
    # Conformal Transform Eq
    xyz_after = translation + (1 + scale) * np.dot(rotation, xyz_before)
    # Convert Vector to Separate Variables
    xtrans = float(xyz_after[0])
    ytrans = float(xyz_after[1])
    ztrans = float(xyz_after[2])
    return xtrans, ytrans, ztrans


def conform14(x, y, z, to_epoch, trans):
    """
    Performs a Helmert 14 Parameter Conformal Transformation using Cartesian point co-ordinates
    and a predefined transformation object. The transformation parameters are projected from
    the transformation objects reference epoch to a specified epoch.
    :param x: Cartesian X (m)
    :param y: Cartesian Y (m)
    :param z: Cartesian Z (m)
    :param to_epoch: Epoch co-ordinate transformation is performed at (datetime.date Object)
    :param trans: Transformation Object
    :return: Cartesian X, Y, Z co-ordinates transformed using Transformation parameters at desired epoch
    """
    if type(trans) != Transformation:
        raise ValueError('trans must be a Transformation Object')
    if type(to_epoch) != datetime.date:
        raise ValueError('to_epoch must be a datetime.date Object')
    # Calculate 7 Parameters from 14 Parameter Transformation Object
    timetrans = trans + to_epoch
    # Perform Transformation
    xtrans, ytrans, ztrans = conform7(x, y, z, timetrans)
    return xtrans, ytrans, ztrans


def mga94_to_mga2020(zone, east, north, ell_ht=False):
    """
    Performs conformal transformation of Map Grid of Australia 1994 to Map Grid of Australia 2020 Coordinates
    using the GDA2020 Tech Manual v1.2 7 parameter similarity transformation parameters
    :param zone: Zone Number - 1 to 60
    :param east: Easting (m, within 3330km of Central Meridian)
    :param north: Northing (m, 0 to 10,000,000m)
    :param ell_ht: Ellipsoid Height (m) (optional)
    :return: MGA2020 Zone, Easting, Northing and Ellipsoid Height (if none provided, returns 0)
    """
    lat, lon, psf, gridconv = grid2geo(zone, east, north)
    if ell_ht is False:
        ell_ht_in = 0
    else:
        ell_ht_in = ell_ht
    x94, y94, z94 = llh2xyz(lat, lon, ell_ht_in)
    x20, y20, z20 = conform7(x94, y94, z94, gda94_to_gda2020)
    lat, lon, ell_ht_out = xyz2llh(x20, y20, z20)
    if ell_ht is False:
        ell_ht_out = 0
    hemisphere, zone20, east20, north20, psf, gridconv = geo2grid(lat, lon)
    return zone20, east20, north20, round(ell_ht_out, 4)


def mga2020_to_mga94(zone, east, north, ell_ht=False):
    """
    Performs conformal transformation of Map Grid of Australia 2020 to Map Grid of Australia 1994 Coordinates
    using the reverse form of the GDA2020 Tech Manual v1.2 7 parameter similarity transformation parameters
    :param zone: Zone Number - 1 to 60
    :param east: Easting (m, within 3330km of Central Meridian)
    :param north: Northing (m, 0 to 10,000,000m)
    :param ell_ht: Ellipsoid Height (m) (optional)
    :return: MGA1994 Zone, Easting, Northing and Ellipsoid Height (if none provided, returns 0)
    """
    lat, lon, psf, gridconv = grid2geo(zone, east, north)
    if ell_ht is False:
        ell_ht_in = 0
    else:
        ell_ht_in = ell_ht
    x94, y94, z94 = llh2xyz(lat, lon, ell_ht_in)
    x20, y20, z20 = conform7(x94, y94, z94, -gda94_to_gda2020)
    lat, lon, ell_ht_out = xyz2llh(x20, y20, z20)
    if ell_ht is False:
        ell_ht_out = 0
    hemisphere, zone20, east20, north20, psf, gridconv = geo2grid(lat, lon)
    return zone20, east20, north20, round(ell_ht_out, 4)


def atrf2014_to_gda2020(x, y, z, epoch_from):
    """
    Transforms Cartesian (x, y, z) Coordinates in terms of the Australian Terrestrial Reference Frame (ATRF) at
    a specified epoch to coordinates in terms of Geocentric Datum of Australia 2020 (GDA2020 - reference epoch 2020.0)
    :param x: ATRF Cartesian X Coordinate (m)
    :param y: ATRF Cartesian Y Coordinate (m)
    :param z: ATRF Cartesian Z Coordinate (m)
    :param epoch_from: ATRF Coordinate Epoch (datetime.date Object)
    :return: Cartesian X, Y, Z Coordinates in terms of GDA2020
    """
    return conform14(x, y, z, epoch_from, atrf_gda2020)


def gda2020_to_atrf2014(x, y, z, epoch_to):
    """
    Transforms Cartesian (x, y, z) Coordinates in terms of Geocentric Datum of Australia 2020
    (GDA2020 - reference epoch 2020.0) to coordinates in terms of the Australian Terrestrial Reference Frame (ATRF) at
    a specified epoch
    :param x: GDA2020 Cartesian X Coordinate (m)
    :param y: GDA2020 Cartesian Y Coordinate (m)
    :param z: GDA2020 Cartesian Z Coordinate (m)
    :param epoch_to: ATRF Coordinate Epoch (datetime.date Object)
    :return: Cartesian X, Y, Z Coordinate in terms of ATRF at the specified Epoch
    """
    return conform14(x, y, z, epoch_to, -atrf_gda2020)
